@@ -1,5 +1,5 @@
 /*
- * $RCSId: xc/lib/fontconfig/src/fcdir.c,v 1.9 2002/08/31 22:17:32 keithp Exp $
+ * fontconfig/src/fcdir.c
  *
  * Copyright © 2000 Keith Packard
  *
@@ -13,9 +13,9 @@
  * representations about the suitability of this software for any purpose.  It
  * is provided "as is" without express or implied warranty.
  *
- * KEITH PACKARD DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
+ * THE AUTHOR(S) DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
  * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
- * EVENT SHALL KEITH PACKARD BE LIABLE FOR ANY SPECIAL, INDIRECT OR
+ * EVENT SHALL THE AUTHOR(S) BE LIABLE FOR ANY SPECIAL, INDIRECT OR
  * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
  * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
  * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
@@ -30,31 +30,22 @@ FcFileIsDir (const FcChar8 *file)
 {
     struct stat	    statb;
 
-    if (stat ((const char *) file, &statb) != 0)
+    if (FcStat ((const char *) file, &statb) != 0)
 	return FcFalse;
     return S_ISDIR(statb.st_mode);
 }
 
-FcBool
-FcFileScanConfig (FcFontSet	*set,
-		  FcStrSet	*dirs,
-		  FcGlobalCache *cache,
-		  FcBlanks	*blanks,
-		  const FcChar8	*file,
-		  FcBool	force,
-		  FcConfig	*config)
+static FcBool
+FcFileScanFontConfig (FcFontSet		*set,
+		      FcBlanks		*blanks,
+		      const FcChar8	*file,
+		      FcConfig		*config)
 {
-    int			id;
-    FcPattern		*font;
-    FcBool		ret = FcTrue;
-    int			count = 0;
+    FcPattern	*font;
+    FcBool	ret = FcTrue;
+    int		id;
+    int		count = 0;
     
-    if (config && !FcConfigAcceptFilename (config, file))
-	return FcTrue;
-
-    if (FcFileIsDir (file))
-	return FcStrSetAdd (dirs, file);
-
     id = 0;
     do
     {
@@ -70,15 +61,31 @@ FcFileScanConfig (FcFontSet	*set,
 	font = FcFreeTypeQuery (file, id, blanks, &count);
 	if (FcDebug () & FC_DBG_SCAN)
 	    printf ("done\n");
+
+	/*
+	 * Edit pattern with user-defined rules
+	 */
+	if (font && config && !FcConfigSubstituteWithPat (config, font, NULL, FcMatchScan))
+	{
+	    FcPatternDestroy (font);
+	    font = NULL;
+	    ret = FcFalse;
+	}
+
 	/*
 	 * Add the font
 	 */
 	if (font && (!config || FcConfigAcceptFont (config, font)))
 	{
+	    if (FcDebug() & FC_DBG_SCANV)
+	    {
+		printf ("Final font pattern:\n");
+		FcPatternPrint (font);
+	    }
 	    if (!FcFontSetAdd (set, font))
 	    {
 		FcPatternDestroy (font);
-		font = 0;
+		font = NULL;
 		ret = FcFalse;
 	    }
 	}
@@ -89,66 +96,70 @@ FcFileScanConfig (FcFontSet	*set,
     return ret;
 }
 
+FcBool
+FcFileScanConfig (FcFontSet	*set,
+		  FcStrSet	*dirs,
+		  FcBlanks	*blanks,
+		  const FcChar8	*file,
+		  FcConfig	*config)
+{
+    if (FcFileIsDir (file))
+	return FcStrSetAdd (dirs, file);
+    else
+	return FcFileScanFontConfig (set, blanks, file, config);
+}
+
 FcBool
 FcFileScan (FcFontSet	    *set,
 	    FcStrSet	    *dirs,
-	    FcGlobalCache   *cache,
+	    FcFileCache	    *cache, /* XXX unused */
 	    FcBlanks	    *blanks,
 	    const FcChar8   *file,
 	    FcBool	    force)
 {
-    return FcFileScanConfig (set, dirs, cache, blanks, file, force, 0);
+    return FcFileScanConfig (set, dirs, blanks, file, NULL);
 }
 
 /*
- * Scan 'dir', adding font files to 'set' and
- * subdirectories to 'dirs'
+ * Strcmp helper that takes pointers to pointers, copied from qsort(3) manpage
  */
+static int
+cmpstringp(const void *p1, const void *p2)
+{
+    return strcmp(* (char **) p1, * (char **) p2);
+}
 
 FcBool
 FcDirScanConfig (FcFontSet	*set,
 		 FcStrSet	*dirs,
-		 FcGlobalCache  *cache,
 		 FcBlanks	*blanks,
-		 const FcChar8  *dir,
-		 FcBool		force,
+		 const FcChar8	*dir,
+		 FcBool		force, /* XXX unused */
 		 FcConfig	*config)
 {
     DIR			*d;
     struct dirent	*e;
+    FcStrSet		*files;
     FcChar8		*file;
-    const FcChar8	*d_can = 0;
     FcChar8		*base;
     FcBool		ret = FcTrue;
-    FcFontSet		*tmpSet;
     int			i;
 
-    if (config && !FcConfigAcceptFilename (config, dir))
+    if (!force)
+	return FcFalse;
+
+    if (!set && !dirs)
 	return FcTrue;
 
-    if (config)
-	d_can = FcConfigNormalizeFontDir (config, dir);
-    if (d_can)
-	dir = d_can;
+    if (!blanks)
+	blanks = FcConfigGetBlanks (config);
 
-    if (!force)
-    {
-	/*
-	 * Check ~/.fonts.cache-<version> file
-	 */
-	if (cache && FcGlobalCacheReadDir (set, dirs, cache, (char *)dir, config))
-	    return FcTrue;
-
-	if (FcDirCacheValid (dir) && 
-	    FcDirCacheHasCurrentArch (dir) &&
-	    FcDirCacheRead (set, dirs, dir, config))
-	    return FcTrue;
-    }
-    
     /* freed below */
     file = (FcChar8 *) malloc (strlen ((char *) dir) + 1 + FC_MAX_FILE_LEN + 1);
-    if (!file)
-	return FcFalse;
+    if (!file) {
+	ret = FcFalse;
+	goto bail;
+    }
 
     strcpy ((char *) file, (char *) dir);
     strcat ((char *) file, "/");
@@ -160,64 +171,158 @@ FcDirScanConfig (FcFontSet	*set,
     d = opendir ((char *) dir);
     if (!d)
     {
-	free (file);
 	/* Don't complain about missing directories */
-	if (errno == ENOENT)
-	    return FcTrue;
-	return FcFalse;
+	if (errno != ENOENT)
+	    ret = FcFalse;
+	goto bail;
     }
 
-    tmpSet = FcFontSetCreate();
-    if (!tmpSet)
-    {	
-	free (file);
-	return FcFalse;
+    files = FcStrSetCreate ();
+    if (!files)
+    {
+	ret = FcFalse;
+	goto bail1;
     }
-
-    while (ret && (e = readdir (d)))
+    while ((e = readdir (d)))
     {
 	if (e->d_name[0] != '.' && strlen (e->d_name) < FC_MAX_FILE_LEN)
 	{
 	    strcpy ((char *) base, (char *) e->d_name);
-	    ret = FcFileScanConfig (tmpSet, dirs, cache, blanks, file, force, config);
+	    if (!FcStrSetAdd (files, file)) {
+		ret = FcFalse;
+		goto bail2;
+	    }
 	}
     }
-    free (file);
-    closedir (d);
+
     /*
-     * Now that the directory has been scanned,
-     * add the cache entry 
+     * Sort files to make things prettier
      */
-    if (ret && cache)
-	FcGlobalCacheUpdate (cache, dirs, (char *)dir, tmpSet, config);
+    qsort(files->strs, files->num, sizeof(FcChar8 *), cmpstringp);
 
-    for (i = 0; i < tmpSet->nfont; i++)
-	FcFontSetAdd (set, tmpSet->fonts[i]);
-
-    if (tmpSet->fonts)
-    {
-	FcMemFree (FC_MEM_FONTPTR, tmpSet->sfont * sizeof (FcPattern *));
-	free (tmpSet->fonts);
-    }
-    FcMemFree (FC_MEM_FONTSET, sizeof (FcFontSet));
-    free (tmpSet);
-	
+    /*
+     * Scan file files to build font patterns
+     */
+    for (i = 0; i < files->num; i++)
+	FcFileScanConfig (set, dirs, blanks, files->strs[i], config);
+    
+bail2:
+    FcStrSetDestroy (files);
+bail1:
+    closedir (d);
+bail:
     return ret;
 }
 
 FcBool
 FcDirScan (FcFontSet	    *set,
 	   FcStrSet	    *dirs,
-	   FcGlobalCache    *cache,
+	   FcFileCache	    *cache, /* XXX unused */
 	   FcBlanks	    *blanks,
 	   const FcChar8    *dir,
-	   FcBool	    force)
+	   FcBool	    force /* XXX unused */)
 {
-    return FcDirScanConfig (set, dirs, cache, blanks, dir, force, 0);
+    if (cache || !force)
+	return FcFalse;
+
+    return FcDirScanConfig (set, dirs, blanks, dir, force, NULL);
+}
+
+/*
+ * Scan the specified directory and construct a cache of its contents
+ */
+FcCache *
+FcDirCacheScan (const FcChar8 *dir, FcConfig *config)
+{
+    FcStrSet		*dirs;
+    FcBool		ret = FcTrue;
+    FcFontSet		*set;
+    FcCache		*cache = NULL;
+    struct stat		dir_stat;
+
+    if (FcDebug () & FC_DBG_FONTSET)
+	printf ("cache scan dir %s\n", dir);
+
+    if (FcStat ((char *) dir, &dir_stat) < 0)
+    {
+	if (errno != ENOENT)
+	    ret = FcFalse;
+	goto bail;
+    }
+
+    set = FcFontSetCreate();
+    if (!set)
+    {
+	ret = FcFalse;
+	goto bail;
+    }
+
+    dirs = FcStrSetCreate ();
+    if (!dirs)
+    {
+	ret = FcFalse;
+	goto bail1;
+    }
+
+    /*
+     * Scan the dir
+     */
+    if (!FcDirScanConfig (set, dirs, NULL, dir, FcTrue, config))
+    {
+	ret = FcFalse;
+	goto bail2;
+    }
+
+    /*
+     * Build the cache object
+     */
+    cache = FcDirCacheBuild (set, dir, &dir_stat, dirs);
+    if (!cache)
+    {
+	ret = FcFalse;
+	goto bail2;
+    }
+    
+    /*
+     * Write out the cache file, ignoring any troubles
+     */
+    FcDirCacheWrite (cache, config);
+    
+ bail2:
+    FcStrSetDestroy (dirs);
+ bail1:
+    FcFontSetDestroy (set);
+ bail:
+    return cache;
+}
+
+/*
+ * Read (or construct) the cache for a directory
+ */
+FcCache *
+FcDirCacheRead (const FcChar8 *dir, FcBool force, FcConfig *config)
+{
+    FcCache		*cache = NULL;
+
+    if (config && !FcConfigAcceptFilename (config, dir))
+	return NULL;
+
+    /* Try to use existing cache file */
+    if (!force)
+	cache = FcDirCacheLoad (dir, config, NULL);
+    
+    /* Not using existing cache file, construct new cache */
+    if (!cache)
+	cache = FcDirCacheScan (dir, config);
+    
+    return cache;
 }
 
 FcBool
 FcDirSave (FcFontSet *set, FcStrSet * dirs, const FcChar8 *dir)
 {
-    return FcDirCacheWrite (set, dirs, dir);
+    return FcFalse; /* XXX deprecated */
 }
+#define __fcdir__
+#include "fcaliastail.h"
+#undef __fcdir__
